@@ -12,7 +12,6 @@
 #include <stdlib.h>
 #include <string.h>
 
-#include "soup-coding.h"
 #include "soup-connection.h"
 #include "soup-message.h"
 #include "soup-message-private.h"
@@ -68,7 +67,7 @@ typedef struct {
 	goffset               write_length;
 	goffset               written;
 
-	guint read_tag, write_tag, err_tag;
+	guint read_tag, write_tag, err_tag, tls_signal_id;
 	GSource *unpause_source;
 
 	SoupMessageGetHeadersFn   get_headers_cb;
@@ -102,6 +101,8 @@ soup_message_io_cleanup (SoupMessage *msg)
 		return;
 	priv->io_data = NULL;
 
+	if (io->tls_signal_id)
+		g_signal_handler_disconnect (io->sock, io->tls_signal_id);
 	if (io->sock)
 		g_object_unref (io->sock);
 	if (io->item)
@@ -185,7 +186,7 @@ io_error (SoupSocket *sock, SoupMessage *msg, GError *error)
 	SoupMessagePrivate *priv = SOUP_MESSAGE_GET_PRIVATE (msg);
 	SoupMessageIOData *io = priv->io_data;
 
-	if (error && error->domain == SOUP_SSL_ERROR) {
+	if (error && error->domain == G_TLS_ERROR) {
 		soup_message_set_status_full (msg,
 					      SOUP_STATUS_SSL_FAILED,
 					      error->message);
@@ -352,25 +353,60 @@ read_metadata (SoupMessage *msg, gboolean to_blank)
 		}
 	}
 
-	if (soup_socket_is_ssl (io->sock)) {
-		gboolean trusted_certificate;
-
-		g_object_get (io->sock,
-			      SOUP_SOCKET_TRUSTED_CERTIFICATE, &trusted_certificate,
-			      NULL);
-
-		if (trusted_certificate)
-			soup_message_set_flags (msg, priv->msg_flags | SOUP_MESSAGE_CERTIFICATE_TRUSTED);
-	}
-
 	return TRUE;
 }
 
+static SoupBuffer *
+content_decode_one (SoupBuffer *buf, GConverter *converter, GError **error)
+{
+	gsize outbuf_length, outbuf_used, outbuf_cur, input_used, input_cur;
+	char *outbuf;
+	GConverterResult result;
+
+	outbuf_length = MAX (buf->length * 2, 1024);
+	outbuf = g_malloc (outbuf_length);
+	outbuf_cur = input_cur = 0;
+
+	do {
+		result = g_converter_convert (
+			converter,
+			buf->data + input_cur, buf->length - input_cur,
+			outbuf + outbuf_cur, outbuf_length - outbuf_cur,
+			0, &input_used, &outbuf_used, error);
+		input_cur += input_used;
+		outbuf_cur += outbuf_used;
+
+		if (g_error_matches (*error, G_IO_ERROR, G_IO_ERROR_NO_SPACE) ||
+		    (!*error && outbuf_cur == outbuf_length)) {
+			g_clear_error (error);
+			outbuf_length *= 2;
+			outbuf = g_realloc (outbuf, outbuf_length);
+		} else if (*error) {
+			/* GZlibDecompressor can't ever return
+			 * G_IO_ERROR_PARTIAL_INPUT unless we pass it
+			 * input_length = 0, which we don't. Other
+			 * converters might of course, so eventually
+			 * this code needs to be rewritten to deal
+			 * with that.
+			 */
+			g_free (outbuf);
+			return NULL;
+		}
+	} while (input_cur < buf->length && result != G_CONVERTER_FINISHED);
+
+	if (outbuf_cur)
+		return soup_buffer_new (SOUP_MEMORY_TAKE, outbuf, outbuf_cur);
+	else {
+		g_free (outbuf);
+		return NULL;
+	}
+}
+
 static SoupBuffer *
 content_decode (SoupMessage *msg, SoupBuffer *buf)
 {
 	SoupMessagePrivate *priv = SOUP_MESSAGE_GET_PRIVATE (msg);
-	SoupCoding *decoder;
+	GConverter *decoder;
 	SoupBuffer *decoded;
 	GError *error = NULL;
 	GSList *d;
@@ -378,10 +414,9 @@ content_decode (SoupMessage *msg, SoupBuffer *buf)
 	for (d = priv->decoders; d; d = d->next) {
 		decoder = d->data;
 
-		decoded = soup_coding_apply (decoder, buf->data, buf->length,
-					     FALSE, &error);
+		decoded = content_decode_one (buf, decoder, &error);
 		if (error) {
-			if (g_error_matches (error, SOUP_CODING_ERROR, SOUP_CODING_ERROR_INTERNAL_ERROR))
+			if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_FAILED))
 				g_warning ("Content-Decoding error: %s\n", error->message);
 			g_error_free (error);
 
@@ -1031,6 +1066,25 @@ io_read (SoupSocket *sock, SoupMessage *msg)
 	goto read_more;
 }
 
+static void
+socket_tls_certificate_changed (GObject *sock, GParamSpec *pspec,
+				gpointer msg)
+{
+	GTlsCertificate *certificate;
+	GTlsCertificateFlags errors;
+
+	g_object_get (sock,
+		      SOUP_SOCKET_TLS_CERTIFICATE, &certificate,
+		      SOUP_SOCKET_TLS_ERRORS, &errors,
+		      NULL);
+	g_object_set (msg,
+		      SOUP_MESSAGE_TLS_CERTIFICATE, certificate,
+		      SOUP_MESSAGE_TLS_ERRORS, errors,
+		      NULL);
+	if (certificate)
+		g_object_unref (certificate);
+}
+
 static SoupMessageIOData *
 new_iostate (SoupMessage *msg, SoupSocket *sock, SoupMessageIOMode mode,
 	     SoupMessageGetHeadersFn get_headers_cb,
@@ -1064,6 +1118,11 @@ new_iostate (SoupMessage *msg, SoupSocket *sock, SoupMessageIOMode mode,
 	io->read_state  = SOUP_MESSAGE_IO_STATE_NOT_STARTED;
 	io->write_state = SOUP_MESSAGE_IO_STATE_NOT_STARTED;
 
+	if (soup_socket_is_ssl (io->sock)) {
+		io->tls_signal_id = g_signal_connect (io->sock, "notify::tls-certificate",
+						      G_CALLBACK (socket_tls_certificate_changed), msg);
+	}
+
 	if (priv->io_data)
 		soup_message_io_cleanup (msg);
 	priv->io_data = io;
